"""
Farm Optimization and eXtended yield Evaluation Software
    
"""

from .core import WindFarm, Turbine  # noqa: F401
from .models import ModelBook  # noqa: F401
from .data import (
    parse_Pct_file_name,
    parse_Pct_two_files,
    FARM,
    STATES,
    PCTCURVE,
    StaticData,
)  # noqa: F401

from . import algorithms  # noqa: F401
from . import models  # noqa: F401
from . import input  # noqa: F401
from . import output  # noqa: F401
from . import utils  # noqa: F401

try:
    from importlib.resources import files

    __version__ = files(__package__).joinpath("VERSION").read_text()
except ImportError:
    from importlib.resources import read_text

    __version__ = read_text(__package__, "VERSION")



import numpy as np

from foxes.models.wake_models.axisymmetric import AxisymmetricWakeModel
from foxes.utils.two_circles import calc_area
import foxes.variables as FV
import foxes.constants as FC

from .centre import PartialCentre


class PartialAxiwake(PartialCentre):
    """
    Partial wake calculation for axial wake models.

    The basic idea is that the x-dependent part of
    the wake model is evaluated only once, and the radial
    part then for `n` radii that cover the target rotor discs.

    The latter results are then weighted according to the overlap
    of radial wake circle area deltas and the target rotor disc area.

    Attributes
    ----------
    n: int
        The number of radial evaluation points

    :group: models.partial_wakes

    """

    def __init__(self, n=6):
        """
        Constructor.

        Parameters
        ----------
        n: int
            The number of radial evaluation points

        """
        super().__init__()
        self.n = n

    def __repr__(self):
        return f"{type(self).__name__}(n={self.n})"

    def check_wmodel(self, wmodel, error=True):
        """
        Checks the wake model type

        Parameters
        ----------
        wmodel: foxes.core.WakeModel
            The wake model to be tested
        error: bool
            Flag for raising TypeError

        Returns
        -------
        chk: bool
            True if wake model is compatible

        """
        if not isinstance(wmodel, AxisymmetricWakeModel):
            if error:
                raise TypeError(
                    f"Partial wakes '{self.name}': Cannot be applied to wake model '{wmodel.name}', since not an AxisymmetricWakeModel"
                )
            return False
        return True

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_deltas,
        wmodel,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)
        wmodel: foxes.core.WakeModel
            The wake model

        """
        # check:
        self.check_wmodel(wmodel, error=True)

        # prepare:
        n_states = mdata.n_states
        n_targets = tdata.n_targets

        # get D:
        D = self.get_data(
            FV.D,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )

        # calc coordinates to rotor centres:
        wcoos = algo.wake_frame.get_wake_coos(algo, mdata, fdata, tdata, downwind_index)

        # prepare x and r coordinates:
        x = np.round(wcoos[..., 0, 0], 12)
        n = wcoos[..., 0, 1:3]
        R = np.linalg.norm(n, axis=-1)
        r = np.zeros((n_states, n_targets, self.n), dtype=FC.DTYPE)
        del wcoos

        # prepare circle section area calculation:
        A = np.zeros((n_states, n_targets, self.n), dtype=FC.DTYPE)
        weights = np.zeros_like(A)

        # get normalized 2D vector between rotor and wake centres:
        sel = R > 0
        if np.any(sel):
            n[sel] /= R[sel][:, None]
        if np.any(~sel):
            n[:, :, 0][~sel] = 1

        # case wake centre outside rotor disk:
        sel = (x > 0) & (R > D / 2)
        if np.any(sel):
            n_sel = np.sum(sel)
            Rsel = np.zeros((n_sel, self.n + 1), dtype=FC.DTYPE)
            Rsel[:] = R[sel][:, None]
            Dsel = D[sel][:, None]

            # equal delta R2:
            R1 = np.zeros((n_sel, self.n + 1), dtype=FC.DTYPE)
            R1[:] = Dsel / 2
            steps = np.linspace(0.0, 1.0, self.n + 1, endpoint=True) - 0.5
            R2 = np.zeros_like(R1)
            R2[:] = Rsel + Dsel * steps[None, :]
            r[sel] = 0.5 * (R2[:, 1:] + R2[:, :-1])

            hA = calc_area(R1, R2, Rsel)
            hA = hA[:, 1:] - hA[:, :-1] + 1e-15

            weights[sel] = hA / np.sum(hA, axis=-1)[:, None]
            del hA, Rsel, Dsel, R1, R2

        # case wake centre inside rotor disk:
        sel = (x > 0) & (R < D / 2)
        if np.any(sel):
            n_sel = np.sum(sel)
            Rsel = np.zeros((n_sel, self.n + 1), dtype=FC.DTYPE)
            Rsel[:] = R[sel][:, None]
            Dsel = D[sel][:, None]

            # equal delta R2:
            R1 = np.zeros((n_sel, self.n + 1), dtype=FC.DTYPE)
            R1[:, 1:] = Dsel / 2
            R2 = np.zeros_like(R1)
            # R2[:, 1:] = Rsel[:, :-1] + Dsel/2
            # R2[:]    *= np.linspace(0., 1, self.n + 1, endpoint=True)[None, :]
            R2[:, 1:] = (Rsel[:, :-1] + Dsel / 2) / (self.n - 0.5)
            R2[:, 1:] *= (
                0.5 + np.linspace(0.0, self.n - 1, self.n, endpoint=True)[None, :]
            )
            hr = 0.5 * (R2[:, 1:] + R2[:, :-1])
            hr[:, 0] = 0.0
            r[sel] = hr

            hA = calc_area(R1, R2, Rsel)
            hA = hA[:, 1:] - hA[:, :-1]
            weights[sel] = hA / np.sum(hA, axis=-1)[:, None]
            del hA, hr, Rsel, Dsel, R1, R2

        # evaluate wake model:
        wdeltas, st_sel = wmodel.calc_wakes_x_r(
            algo, mdata, fdata, tdata, downwind_index, x, r
        )

        for v, wdel in wdeltas.items():
            d = np.einsum("sn,sn->s", wdel, weights[st_sel])

            try:
                superp = wmodel.superp[v]
            except KeyError:
                raise KeyError(
                    f"Model '{self.name}': Missing wake superposition entry for variable '{v}' in wake model '{wmodel.name}', found {sorted(list(wmodel.superp.keys()))}"
                )

            wake_deltas[v] = superp.add_wake(
                algo,
                mdata,
                fdata,
                tdata,
                downwind_index,
                st_sel,
                v,
                wake_deltas[v],
                d[:, None],
            )



import numpy as np


def cubic_roots(a0, a1, a2, a3=None):
    """
    Calculate real roots of polynomials of degree 3.

    Convention:
    f(x) = a[3]*x**3 + a[2]*x**2 + a[1]*x + a[0]

    In contrast to numpy's "root" function
    this works fast for an array of polynomials,
    so you spare yourself looping over them.

    Source: https://github.com/opencv/opencv/blob/master/modules/calib3d/src/polynom_solver.cpp

    Parameters
    ----------
    a0: numpy.ndarray
        The coefficients a[0]
    a1: numpy.ndarray
        The coefficients a[1]
    a2: numpy.ndarray
        The coefficients a[2]
    a3: numpy.ndarray
        The coefficients a[3], or None for ones

    Returns
    -------
    roots: numpy.ndarray
        The real roots of the polynomial,
        shape: (n_a0, 3). If one root only
        the two last columns will be np.nan

    :group: utils

    """

    N = len(a0)
    out = np.full([N, 3], np.nan)

    # Calculate the normalized form x^3 + a2 * x^2 + a1 * x + a0 = 0
    b_a = a2 if a3 is None else a2 / a3
    b_a2 = b_a * b_a
    c_a = a1 if a3 is None else a1 / a3
    d_a = a0 if a3 is None else a0 / a3

    # Solve the cubic equation
    Q = (3 * c_a - b_a2) / 9
    R = (9 * b_a * c_a - 27 * d_a - 2 * b_a * b_a2) / 54
    Q3 = Q * Q * Q
    D = Q3 + R * R
    b_a_3 = (1.0 / 3.0) * b_a

    sel = Q == 0.0
    if np.any(sel):
        o = out[sel, 0]

        sel2 = R == 0.0
        if np.any(sel2):
            o[sel2] = -b_a_3[sel][sel2]

        if np.any(~sel2):
            o[~sel2] = np.pow(2 * R[sel][~sel2], 1 / 3.0) - b_a_3[sel][~sel2]

        out[sel, 0] = o

    sel = D <= 0.0
    if np.any(sel):
        # Three real roots
        theta = np.arccos(R[sel] / np.sqrt(-Q3[sel]))
        sqrt_Q = np.sqrt(-Q[sel])

        out[sel, 0] = 2 * sqrt_Q * np.cos(theta / 3.0) - b_a_3[sel]
        out[sel, 1] = 2 * sqrt_Q * np.cos((theta + 2 * np.pi) / 3.0) - b_a_3[sel]
        out[sel, 2] = 2 * sqrt_Q * np.cos((theta + 4 * np.pi) / 3.0) - b_a_3[sel]

    return out


def test_cubic_roots(roots, a0, a1, a2, a3=None, tol=1.0e-12):
    """
    Test the cubic roots results

    Parameters
    ----------
    roots: numpy.ndarray
        The roots to test, shape: (n_a0, 3)
    a0: numpy.ndarray
        The coefficients a[0]
    a1: numpy.ndarray
        The coefficients a[1]
    a2: numpy.ndarray
        The coefficients a[2]
    a3: numpy.ndarray
        The coefficients a[3], or None for ones

    """

    N = len(a0)
    for n in range(N):
        c0 = a0[n]
        c1 = a1[n]
        c2 = a2[n]
        c3 = a3[n]

        print(f"Polynomial {n}: a = {(c0,c1,c2,c3)}")

        rts = np.unique(roots[n])
        rts = rts[~np.isnan(rts)]

        for x in rts:
            f = c0 + c1 * x + c2 * x**2 + c3 * x**3
            ok = np.abs(f) <= tol

            print(f"  root x = {x}: f(x) = {f}     {'OK' if ok else 'FAILED'}")

            if not ok:
                raise Exception("NOT OK!")

        if len(rts) == 0:
            print("  no real roots.")


if __name__ == "__main__":
    N = 100
    a0 = np.random.uniform(-10.0, 10.0, N)
    a1 = np.random.uniform(-10.0, 10.0, N)
    a2 = np.random.uniform(-10.0, 10.0, N)
    a3 = np.random.uniform(1.0, 10.0, N)

    roots = cubic_roots(a0, a1, a2, a3)

    test_cubic_roots(roots, a0, a1, a2, a3)



import numpy as np

from foxes.core import VerticalProfile
from foxes.utils import abl
import foxes.variables as FV


class ShearedProfile(VerticalProfile):
    """
    A wind shear profile, based on a shear exponent.

    :group: models.vertical_profiles

    """

    def input_vars(self):
        """
        The input variables needed for the profile
        calculation.

        Returns
        -------
        vars: list of str
            The variable names

        """
        return [FV.WS, FV.H, FV.SHEAR]

    def calculate(self, data, heights):
        """
        Run the profile calculation.

        Parameters
        ----------
        data: dict
            The input data
        heights: numpy.ndarray
            The evaluation heights

        Returns
        -------
        results: numpy.ndarray
            The profile results, same
            shape as heights

        """

        ws = np.zeros_like(heights)
        ws[:] = data[FV.WS]

        h0 = np.zeros_like(heights)
        h0[:] = data[FV.H]

        shear = np.zeros_like(heights)
        shear[:] = data[FV.SHEAR]

        out = np.zeros_like(heights)
        out[:] = abl.sheared.calc_ws(heights, h0, ws, shear)

        return out



"""
Verctical profile models.
"""

from .uniform import UniformProfile
from .abl_log_neutral_ws import ABLLogNeutralWsProfile
from .abl_log_stable_ws import ABLLogStableWsProfile
from .abl_log_unstable_ws import ABLLogUnstableWsProfile
from .abl_log_ws import ABLLogWsProfile
from .sheared_ws import ShearedProfile
from .data_profile import DataProfile



import numpy as np
import matplotlib.pyplot as plt

from .circle import Circle
from .polygon import ClosedPolygon
from .area_geometry import AreaUnion

if __name__ == "__main__":
    boundary = (
        ClosedPolygon(
            np.array([[0, 0], [0, 1200], [1000, 800], [900, -200]], dtype=np.float64)
        )
        + ClosedPolygon(
            np.array([[500, 0], [500, 1500], [1000, 1500], [1000, 0]], dtype=np.float64)
        )
        - Circle([-100.0, -100.0], 700)
    )

    fig, ax = plt.subplots()
    boundary.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)
    quit()

    centres = [np.array([7.0, 8.0])]
    radii = [0.5]
    N = 500

    plist = [
        np.array([[1.0, 1.0], [1.3, 6], [5.8, 6.2], [6.5, 0.8]]),
        np.array([[1.5, 1.5], [1.5, 8], [2.5, 8], [2.5, 1.5]]),
    ]

    circles = [Circle(centres[i], radii[i]) for i in range(len(centres))]
    polygons = [ClosedPolygon(plist[i]) for i in range(len(plist))]

    fig, ax = plt.subplots()
    g = AreaUnion(circles + polygons)
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = AreaUnion(circles + polygons)
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = AreaUnion(circles + polygons).inverse()
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = AreaUnion(circles + polygons).inverse()
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = polygons[0] + circles[0]
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)



import numpy as np
import pandas as pd
import xarray as xr
import matplotlib.pyplot as plt


def exec_python(s, indicator="%", newline=";", globals=globals(), locals={}):
    """
    Executes strings that start with the
    indicator as python commands, returns one value

    Example:
    s = "%p%N=10;p=np.zeros((N,3));p[:,0]=50;p[:,1]=np.linspace(0,7000,N);p[:,2]=119"

    Parameters
    ----------
    s: list, dict or object
        The source to by analyzed
    indicator: str
        The indicator that trigger python evaluation
    newline: str
        The new line indicator
    globals: dict
        The global namespace
    locals: dict
        The local namespace

    Returns
    -------
    out: list, dict or object
        The same structure, but all python
        strings evaluated

    :group: utils

    """
    if isinstance(s, str):
        L = len(indicator)
        if len(s) > L and s[:L] == indicator:
            a = s[L:]
            if not indicator in a:
                exec(a, globals, locals)
            else:
                ilist = a.split(indicator)
                if len(ilist) != 2:
                    raise ValueError(
                        f"Expecting at most 2 occurences of '{indicator}', found {len(ilist)}: {s}"
                    )
                v, b = ilist
                for c in b.split(newline):
                    exec(c, globals, locals)
                return locals[v]
    elif isinstance(s, list):
        return [exec_python(a, indicator) for a in s]
    elif isinstance(s, tuple):
        return tuple(exec_python(list(s), indicator))
    elif isinstance(s, dict):
        return {k: exec_python(a, indicator) for k, a in s.items()}
    return s



import numpy as np

from foxes.core import TurbineModel
import foxes.variables as FV
import foxes.constants as FC
from foxes.utils import cubic_roots


class PowerMask(TurbineModel):
    """
    Invokes a maximal power value.

    This may correspond to turbine derating, if
    the maximal power value is below rated power.
    For higher values, a boost is introduced.

    The model updates the P and CT variables,
    so it is wise to use it after calling the
    turbine type model.

    Attributes
    ----------
    var_ws_P: str
        The wind speed variable for power lookup
    factor_P: float
        The power unit factor, e.g. 1000 for kW
    P_lim: float
        Threshold power delta for boosts
    induction: foxes.core.AxialInductionModel
        The induction model

    :group: models.turbine_models

    """

    def __init__(self, var_ws_P=FV.REWS3, factor_P=1.0e3, P_lim=100, induction="Betz"):
        """
        Constructor.

        Parameters
        ----------
        var_ws_P: str
            The wind speed variable for power lookup
        factor_P: float
            The power unit factor, e.g. 1000 for kW
        P_lim: float
            Threshold power delta for boosts
        induction: foxes.core.AxialInductionModel or str
            The induction model

        """
        super().__init__()

        self.var_ws_P = var_ws_P
        self.factor_P = factor_P
        self.P_lim = P_lim
        self.induction = induction

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        a = f"var_ws_P={self.var_ws_P}, P_lim={self.P_lim}, induction={iname}"
        return f"{type(self).__name__}({a})"

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.P, FV.CT]

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.induction]

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity)

        self._P_rated = []
        for t in algo.farm_controller.turbine_types:
            Pnom = FC.DTYPE(t.P_nominal)
            if np.isnan(Pnom):
                raise ValueError(
                    f"Model '{self.name}': P_nominal is NaN for turbine type '{t.name}'"
                )
            self._P_rated.append(Pnom)
        self._P_rated = np.array(self._P_rated, dtype=FC.DTYPE)

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        # prepare:
        P = fdata[FV.P]
        max_P = fdata[FV.MAX_P]
        P_rated = self._P_rated[None, :]

        # select power entries for which this is active:
        sel = np.zeros((fdata.n_states, fdata.n_turbines), dtype=bool)
        sel[st_sel] = True
        sel = (
            sel
            & ~np.isnan(max_P)
            & (
                ((max_P < P_rated) & (P > max_P))
                | ((max_P > P_rated) & (P > P_rated - self.P_lim))
            )
        )
        if np.any(sel):
            # apply selection:
            max_P = max_P[sel]
            ws = fdata[self.var_ws_P][sel]
            rho = fdata[FV.RHO][sel]
            r = fdata[FV.D][sel] / 2
            P = P[sel]
            ct = fdata[FV.CT][sel]

            # calculate power efficiency e of turbine
            # e is the ratio of the cp derived from the power curve
            # and the theoretical cp from the turbine induction
            cp = P / (0.5 * ws**3 * rho * np.pi * r**2) * self.factor_P
            a = self.induction.ct2a(ct)
            cp_a = 4 * a**3 - 8 * a**2 + 4 * a
            e = cp / cp_a
            del cp, a, cp_a, ct, P

            # calculating new cp for changed power
            cp = max_P / (0.5 * ws**3 * rho * np.pi * r**2) * self.factor_P

            # find roots:
            N = len(cp)
            a3 = np.full(N, 4.0, dtype=FC.DTYPE)
            a2 = np.full(N, -8.0, dtype=FC.DTYPE)
            a1 = np.full(N, 4.0, dtype=FC.DTYPE)
            a0 = -cp / e
            rts = cubic_roots(a0, a1, a2, a3)
            rts[np.isnan(rts)] = np.inf
            rts[rts <= 0.0] = np.inf
            a = np.min(rts, axis=1)
            del a0, a1, a2, a3, rts

            # set results:
            P = fdata[FV.P]
            ct = fdata[FV.CT]
            P[sel] = max_P
            ct[sel] = 4 * a * (1 - a)

        return {FV.P: fdata[FV.P], FV.CT: fdata[FV.CT]}



import numpy as np
from pathlib import Path

from foxes.utils import Dict
import foxes.variables as FV
import foxes.constants as FC

def _read_turbine_outputs(wio_outs, odir, out_dicts, verbosity):
    """ Reads the turbine outputs request """
    if "turbine_outputs" in wio_outs and wio_outs["turbine_outputs"].get("report", True):
        turbine_outputs = Dict(wio_outs["turbine_outputs"], name="turbine_outputs")
        turbine_nc_filename = turbine_outputs.pop("turbine_nc_filename", "turbine_outputs.nc")
        output_variables = turbine_outputs["output_variables"]
        if verbosity > 2:
            print("      Reading turbine_outputs")
            print("        File name:", turbine_nc_filename)
            print("        output_variables:", output_variables)
        
        vmap = Dict(
            power=FV.P,
            rotor_effective_velocity=FV.REWS,
        )
        ivmap = {d: k for k, d in vmap.items()}
        ivmap.update({
            FC.STATE: "time", 
            FC.TURBINE: "turbine",
        })
        
        out_dicts.append(Dict({
                    "output_type": "StateTurbineTable",
                    "farm_results": True,
                    "algo": False,
                    "run_func": "get_dataset",
                    "run_kwargs": dict(
                        variables=[vmap[v] for v in output_variables],
                        name_map=ivmap,
                        to_file=odir/turbine_nc_filename,
                        verbosity=verbosity,
                    ),
                    "output_yaml_update": {
                        "power_table": f"include {turbine_nc_filename}",
                    },
                }, name = "turbine_outputs"))

def _read_flow_field(wio_outs, odir, out_dicts, verbosity):
    """ Reads the flow field request """
    if "flow_field" in wio_outs and wio_outs["flow_field"].get("report", True):
        flow_field = Dict(wio_outs["flow_field"], name="flow_field")
        flow_nc_filename = flow_field.pop("flow_nc_filename", "flow_field.nc")
        output_variables = flow_field.pop("output_variables")
        z_planes = Dict(flow_field.pop("z_planes"), name="z_planes")
        z_sampling = z_planes["z_sampling"]
        xy_sampling = z_planes["xy_sampling"]
        if verbosity > 2:
            print("      Reading flow_field")
            print("        File name       :", flow_nc_filename)
            print("        output_variables:", output_variables)
            print("        z_sampling      :", z_sampling)
            print("        xy_sampling     :", xy_sampling)

        vmap = Dict(
            wind_speed=FV.WS,
            wind_direction=FV.WD,
        )
        
        if xy_sampling == "default":
            out_dicts.append(Dict({
                        "output_type": "SliceData",
                        "farm_results": True,
                        "algo": True,
                        "verbosity_delta": 3,
                        "run_func": "get_states_data_xy",
                        "run_kwargs": dict(
                            resolution=30.,
                            variables=[vmap[v] for v in output_variables],
                            z=None if z_sampling == "hub_height" else z_sampling,
                            to_file=odir/flow_nc_filename,
                            verbosity=verbosity,
                        ),
                        "output_yaml_update": {
                            "flow_field": f"include {flow_nc_filename}",
                        },
                    }, name = "flow_field"))
        else:
            raise NotImplementedError(f"xy_sampling '{xy_sampling}' is not supported (yet)")
            
        
def read_outputs(wio_outs, algo_dict, verbosity):
    """
    Reads the windio outputs

    Parameters
    ----------
    wio_outs: dict
        The windio output data
    algo_dict: dict
        The algorithm dictionary
    verbosity: int
        The verbosity level, 0=silent

    Returns
    -------
    out_dicts: list of dict
        The output dictionaries
    odir: pathlib.Path
        Path to the output folder

    :group: input.windio

    """
    out_dicts = []
    odir = Path(wio_outs.pop("output_folder", "results"))
    odir.mkdir(exist_ok=True, parents=True)
    if verbosity > 2:
        print("  Reading outputs")
        print("    Output folder:", odir)
        print("    Contents:", [k for k in wio_outs.keys()])

    # read turbine_outputs:
    _read_turbine_outputs(wio_outs, odir, out_dicts, verbosity)

    # read flow field:
    _read_flow_field(wio_outs, odir, out_dicts, verbosity)

    return out_dicts, odir



from .model import Model


class GroundModel(Model):
    """
    Base class for ground models.
    """

    def new_farm_wake_deltas(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        wmodel,
        pwake,
    ):
        """
        Creates new initial wake deltas, filled
        with zeros.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        tdata: foxes.core.Data
            The target point data
        wmodel: foxes.core.WakeModel
            The wake model
        pwake: foxes.core.PartialWakesModel
            The partial wakes model

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_turbines, n_tpoints, ...)

        """
        return pwake.new_wake_deltas(algo, mdata, fdata, tdata, wmodel)

    def contribute_to_farm_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_deltas,
        wmodel,
        pwake,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)
        wmodel: foxes.core.WakeModel
            The wake model
        pwake: foxes.core.PartialWakesModel
            The partial wakes model

        """
        pwake.contribute(algo, mdata, fdata, tdata, downwind_index, wake_deltas, wmodel)

    def finalize_farm_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        amb_res,
        rpoint_weights,
        wake_deltas,
        wmodel,
        downwind_index,
        pwake,
    ):
        """
        Updates the wake_deltas at the selected target
        downwind index.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.Data
            The target point data
        amb_res: dict
            The ambient results at the target points
            of all rotors. Key: variable name, value
            np.ndarray of shape:
            (n_states, n_turbines, n_rotor_points)
        rpoint_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: np.ndarray of shape
            (n_states, n_turbines, n_tpoints)
        wmodel: foxes.core.WakeModel
            The wake model
        downwind_index: int
            The index in the downwind order

        Returns
        -------
        final_wake_deltas: dict
            The final wake deltas at the selected downwind
            turbines. Key: variable name, value: np.ndarray
            of shape (n_states, n_rotor_points)

        """
        return pwake.finalize_wakes(
            algo,
            mdata,
            fdata,
            tdata,
            amb_res,
            rpoint_weights,
            wake_deltas,
            wmodel,
            downwind_index,
        )

    def new_point_wake_deltas(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        wmodel,
    ):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        wmodel: foxes.core.WakeModel
            The wake model

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_targets, n_tpoints, ...)

        """
        return wmodel.new_wake_deltas(algo, mdata, fdata, tdata)

    def contribute_to_point_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_deltas,
        wmodel,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)
        wmodel: foxes.core.WakeModel
            The wake model

        """
        wcoos = algo.wake_frame.get_wake_coos(algo, mdata, fdata, tdata, downwind_index)
        wmodel.contribute(algo, mdata, fdata, tdata, downwind_index, wcoos, wake_deltas)

    def finalize_point_wakes(
        self,
        algo,
        mdata,
        fdata,
        amb_results,
        wake_deltas,
        wmodel,
    ):
        """
        Finalize the wake calculation.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        amb_results: dict
            The ambient results, key: variable name str,
            values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)
        wake_deltas: dict
            The wake deltas object at the selected target
            turbines. Key: variable str, value: numpy.ndarray
            with shape (n_states, n_targets, n_tpoints)

        """
        wmodel.finalize_wake_deltas(algo, mdata, fdata, amb_results, wake_deltas)



import numpy as np
import xarray as xr
from abc import abstractmethod
from dask.distributed import progress
from dask.diagnostics import ProgressBar

from .model import Model
from .data import MData, FData, TData
from foxes.utils.runners import DaskRunner
import foxes.constants as FC
import foxes.variables as FV


class DataCalcModel(Model):
    """
    Abstract base class for models with
    that run calculation on xarray Dataset
    data.

    The calculations are run via xarray's
    `apply_ufunc` function, i.e., they run in
    parallel depending on the dask settings.

    For each individual data chunk the `calculate`
    function is called.

    :group: core

    """

    @abstractmethod
    def calculate(self, algo, *data, **parameters):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data: tuple of foxes.core.Data
            The input data
        parameters: dict, optional
            The calculation parameters

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray

        """
        pass

    def _wrap_calc(
        self,
        *ldata,
        algo,
        dvars,
        lvars,
        ldims,
        evars,
        edims,
        edata,
        loop_dims,
        out_vars,
        out_dims,
        calc_pars,
        init_vars,
    ):
        """
        Wrapper that mitigates between apply_ufunc and `calculate`.
        """
        n_prev = len(init_vars)
        if n_prev:
            prev = ldata[:n_prev]
            ldata = ldata[n_prev:]

        # reconstruct original data:
        data = []
        for i, hvars in enumerate(dvars):
            v2l = {v: lvars.index(v) for v in hvars if v in lvars}
            v2e = {v: evars.index(v) for v in hvars if v in evars}

            hdata = {v: ldata[v2l[v]] if v in v2l else edata[v2e[v]] for v in hvars}
            hdims = {v: ldims[v2l[v]] if v in v2l else edims[v2e[v]] for v in hvars}

            if i == 0:
                data.append(MData(data=hdata, dims=hdims, loop_dims=loop_dims))
            elif i == 1:
                data.append(FData(data=hdata, dims=hdims, loop_dims=loop_dims))
            elif i == 2:
                data.append(TData(data=hdata, dims=hdims, loop_dims=loop_dims))
            else:
                raise NotImplementedError(
                    f"Not more than 3 data sets implemented, found {len(dvars)}"
                )

            del hdata, hdims, v2l, v2e

        # deduce output shape:
        oshape = []
        for li, l in enumerate(out_dims):
            for i, dims in enumerate(ldims):
                if l in dims:
                    oshape.append(ldata[i].shape[dims.index(l)])
                    break
            if len(oshape) != li + 1:
                raise ValueError(f"Model '{self.name}': Failed to find loop dimension")

        # add zero output data arrays:
        odims = {v: tuple(out_dims) for v in out_vars}
        odata = {
            v: (
                np.full(oshape, np.nan, dtype=FC.DTYPE)
                if v not in init_vars
                else prev[init_vars.index(v)].copy()
            )
            for v in out_vars
            if v not in data[-1]
        }

        if len(data) == 1:
            data.append(FData(odata, odims, loop_dims))
        else:
            odata.update(data[-1])
            odims.update(data[-1].dims)
            if len(data) == 2:
                data[-1] = FData(odata, odims, loop_dims)
            else:
                data[-1] = TData(odata, odims, loop_dims)
        del odims, odata

        # link chunk state indices from mdata to fdata and tdata:
        if FC.STATE in data[0]:
            for d in data[1:]:
                d[FC.STATE] = data[0][FC.STATE]

        # link weights from mdata to fdata:
        if FV.WEIGHT in data[0]:
            data[1][FV.WEIGHT] = data[0][FV.WEIGHT]
            data[1].dims[FV.WEIGHT] = data[0].dims[FV.WEIGHT]

        # run model calculation:
        self.ensure_variables(algo, *data)
        results = self.calculate(algo, *data, **calc_pars)

        # replace missing results by first input data with matching shape:
        missing = set(out_vars).difference(results.keys())
        if len(missing):
            found = set()
            for v in missing:
                for dta in data:
                    if v in dta and dta[v].shape == tuple(oshape):
                        results[v] = dta[v]
                        found.add(v)
                        break
            missing -= found
            if len(missing):
                raise ValueError(
                    f"Model '{self.name}': Missing results {list(missing)}, expected shape {oshape}"
                )
        del data

        # create output:
        n_vars = len(out_vars)
        data = np.zeros(oshape + [n_vars], dtype=FC.DTYPE)
        for v in out_vars:
            data[..., out_vars.index(v)] = results[v]

        return data

    def run_calculation(
        self,
        algo,
        *data,
        out_vars,
        loop_dims,
        out_core_vars,
        initial_results=None,
        sel=None,
        isel=None,
        **calc_pars,
    ):
        """
        Starts the model calculation in parallel, via
        xarray's `apply_ufunc`.

        Typically this function is called by algorithms.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data: tuple of xarray.Dataset
            The input data
        out_vars: list of str
            The calculation output variables
        loop_dims: array_like of str
            List of the loop dimensions during xarray's
            `apply_ufunc` calculations
        out_core_vars: list of str
            The core dimensions of the output data, use
            `FC.VARS` for variables dimension (required)
        initial_results: xarray.Dataset, optional
            Initial results
        sel: dict, optional
            Selection of loop_dim variable subset values
        isel: dict, optional
            Selection of loop_dim variable subset index values
        calc_pars: dict, optional
            Additional arguments for the `calculate` function

        Returns
        -------
        results: xarray.Dataset
            The calculation results

        """
        # check:
        if not self.initialized:
            raise ValueError(
                f"DataCalcModel '{self.name}': run_calculation called for uninitialized model"
            )

        # prepare:
        loopd = set(loop_dims)

        # extract loop-var dependent and independent data:
        ldata = []
        lvars = []
        ldims = []
        edata = []
        evars = []
        edims = []
        dvars = []
        ivars = []
        idims = []
        if initial_results is not None:
            ds = initial_results
            hvarsl = [v for v, d in ds.items() if len(loopd.intersection(d.dims))]
            ldata += [ds[v] for v in hvarsl]
            idims += [ds[v].dims for v in hvarsl]
            ivars += hvarsl

        for ds in data:
            hvarsl = [v for v, d in ds.items() if len(loopd.intersection(d.dims))]
            ldata += [ds[v] for v in hvarsl]
            ldims += [ds[v].dims for v in hvarsl]
            lvars += hvarsl

            hvarse = [v for v in ds.keys() if v not in hvarsl]
            edata += [ds[v].values for v in hvarse]
            edims += [ds[v].dims for v in hvarse]
            evars += hvarse

            for c, d in ds.coords.items():
                if c in loopd:
                    ldata.append(xr.DataArray(data=d.values, coords={c: d}, dims=[c]))
                    ldims.append((c,))
                    lvars.append(c)
                else:
                    edata.append(d.values)
                    edims.append((c,))
                    evars.append(c)

            dvars.append(list(ds.keys()) + list(ds.coords.keys()))

        # subset selection:
        if sel is not None:
            nldata = []
            for ds in ldata:
                s = {k: v for k, v in sel.items() if k in ds.coords}
                if len(s):
                    nldata.append(ds.sel(s))
            ldata = nldata
            del nldata
        if isel is not None:
            nldata = []
            for ds in ldata:
                s = {k: v for k, v in isel.items() if k in ds.coords}
                if len(s):
                    nldata.append(ds.isel(s))
            ldata = nldata
            del nldata

        # setup dask options:
        dargs = dict(output_sizes={FC.VARS: len(out_vars)})
        if FC.TURBINE in loopd and FC.TURBINE not in ldims.values():
            dargs["output_sizes"][FC.TURBINE] = algo.n_turbines
        if FC.VARS not in out_core_vars:
            raise ValueError(
                f"Model '{self.name}': Expecting '{FC.VARS}' in out_core_vars, got {out_core_vars}"
            )

        # setup arguments for wrapper function:
        out_dims = loop_dims + list(set(out_core_vars).difference([FC.VARS]))
        wargs = dict(
            algo=algo,
            dvars=dvars,
            lvars=lvars,
            ldims=ldims,
            evars=evars,
            edims=edims,
            edata=edata,
            loop_dims=loop_dims,
            out_vars=out_vars,
            out_dims=out_dims,
            calc_pars=calc_pars,
            init_vars=ivars,
        )

        # run parallel computation:
        iidims = [[c for c in d if c not in loopd] for d in idims]
        icdims = [[c for c in d if c not in loopd] for d in ldims]
        results = xr.apply_ufunc(
            self._wrap_calc,
            *ldata,
            input_core_dims=iidims + icdims,
            output_core_dims=[out_core_vars],
            output_dtypes=[FC.DTYPE],
            dask="parallelized",
            dask_gufunc_kwargs=dargs,
            kwargs=wargs,
        )

        # reorganize results Dataset:
        results = results.assign_coords({FC.VARS: out_vars}).to_dataset(dim=FC.VARS)

        if DaskRunner.is_distributed() and len(ProgressBar.active):
            progress(results.persist())

        # update data by calculation results:
        return results.compute()



"""
Wake frame models.
"""

from .rotor_wd import RotorWD
from .streamlines import Streamlines2D
from .timelines import Timelines
from .yawed_wakes import YawedWakes
from .farm_order import FarmOrder
from .seq_dynamic_wakes import SeqDynamicWakes



import numpy as np

XYH = "xyh"
""" The vector (x, y, height)
:group: foxes.variables
"""

FARM = "farm"
""" Wind farm identifier
:group: foxes.constants
"""

STATE = "state"
""" States identifier
:group: foxes.constants
"""

TIME = "time"
""" Time identifier
:group: foxes.constants
"""

TURBINE = "turbine"
""" Wind turbine identifier
:group: foxes.constants
"""

TNAME = "tname"
""" Wind turbine name identifier
:group: foxes.constants
"""

TARGET = "target"
""" Target identifier
:group: foxes.constants
"""

TARGETS = "targets"
""" Targets identifier
:group: foxes.constants
"""

TPOINT = "target_point"
""" Target point identifier
:group: foxes.constants
"""

TPOINTS = "target_points"
""" Points per target identifier
:group: foxes.constants
"""

TWEIGHTS = "tpoint_weights"
""" Target point weights identifier
:group: foxes.constants
"""

POINT = "point"
""" Point identifier
:group: foxes.constants
"""

POINTS = "points"
""" Points identifier
:group: foxes.constants
"""

AMB_TARGET_RESULTS = "amb_target_res"
""" Identifier for ambient target results
:group: foxes.constants
"""


VARS = "vars"
""" Variables identifier
:group: foxes.constants
"""

VALID = "valid"
""" Validity identifier
:group: foxes.constants
"""

TMODELS = "tmodels"
""" Turbine models identifier
:group: foxes.constants
"""

TMODEL_SELS = "tmodel_sels"
"""Selected turbine models identifier
:group: foxes.constants
"""

STATES_SEL = "states_sel"
"""Identifier for states selection
:group: foxes.constants
"""

STATE_TURBINE = "state-turbine"
"""Identifier for state-turbine dimensions
:group: foxes.constants
"""

STATE_TARGET = "state-target"
"""Identifier for state-target dimensions
:group: foxes.constants
"""

STATE_TARGET_TPOINT = "state-target-tpoint"
"""Identifier for state-target-tpoints dimensions
:group: foxes.constants
"""

STATE_SOURCE_ORDERI = "state-source-orderi"
"""Identifier for order index of wake causing turbines
:group: foxes.constants
"""

DTYPE = np.float64
""" Default data type for floats
:group: foxes.constants
"""

ITYPE = np.int64
""" Default data type for int
:group: foxes.constants
"""


KAPPA = 0.41
""" The Van-Karman constant
:group: foxes.constants
"""


W = "W"
""" The unit watt
:group: foxes.constants
"""

kW = "kW"
""" The unit kilo watt
:group: foxes.constants
"""

MW = "MW"
""" The unit mega watt
:group: foxes.constants
"""

GW = "GW"
""" The unit giga watt
:group: foxes.constants
"""

TW = "TW"
""" The unit terra watt
:group: foxes.constants
"""

P_UNITS = {W: 1.0, kW: 1.0e3, MW: 1.0e6, GW: 1.0e9, TW: 1.0e12}
""" Power unit factors relative to watts,
key: unit str, value: factor
:group: foxes.constants
"""


POP = "pop"
""" Population identifier
:group: foxes.constants
"""



import numpy as np

from foxes.models.wake_models.dist_sliced import DistSlicedWakeModel
from foxes.core import Model, WakeK
import foxes.variables as FV
import foxes.constants as FC


class Bastankhah2016Model(Model):
    """
    Common calculations for the wake model and the wake
    frame, such that code repetitions can be avoided.

    Notes
    -----
    Reference:
    "Experimental and theoretical study of wind turbine wakes in yawed conditions"
    Majid Bastankhah, Fernando Porté-Agel
    https://doi.org/10.1017/jfm.2016.595

    Attributes
    ----------
    alpha: float
        model parameter used to determine onset of far wake region
    beta: float
        model parameter used to determine onset of far wake region
    induction: foxes.core.AxialInductionModel or str
        The induction model

    :group: models.wake_models.wind

    """

    MDATA_KEY = "Bastankhah2016Model"
    PARS = "pars"
    CHECK = "check"
    ST_SEL = "st_sel"
    X0 = "x0"

    NEAR = "near"
    R_PC = "r_pc"
    R_PC_S = "r_pc_s"
    AMPL_NEAR = "ampl_near"
    DELTA_NEAR = "delta_near"

    AMPL_FAR = "ampl_far"
    SIGMA_Y_FAR = "sigma_y_far"
    SIGMA_Z_FAR = "sigma_z_far"
    DELTA_FAR = "delta_far"

    def __init__(self, alpha, beta, induction):
        """
        Constructor.

        Parameters
        ----------
        alpha: float
            model parameter used to determine onset of far wake region
        beta: float
            model parameter used to determine onset of far wake region
        induction: foxes.core.AxialInductionModel or str
            The induction model

        """
        super().__init__()
        self.induction = induction
        setattr(self, FV.PA_ALPHA, alpha)
        setattr(self, FV.PA_BETA, beta)

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    @property
    def pars(self):
        """
        Dictionary of the model parameters

        Returns
        -------
        dict :
            Dictionary of the model parameters

        """
        alpha = getattr(self, FV.PA_ALPHA)
        beta = getattr(self, FV.PA_BETA)
        return dict(alpha=alpha, beta=beta, induction=self.induction.name)

    def calc_data(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        gamma,
        k,
    ):
        """
        Calculate common model data, store it in mdata.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        gamma: numpy.ndarray
            The YAWM angles in radiants, shape: (n_states, n_targets)
        k: numpy.ndarray
            The k parameter values, shape: (n_states, n_targets)

        """

        # store parameters:
        out = {self.PARS: self.pars}
        out[self.CHECK] = (
            mdata[FC.STATE][0],
            downwind_index,
            x.shape,
        )

        # get D:
        D = super().get_data(
            FV.D,
            target=FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )

        # get ct:
        ct = super().get_data(
            FV.CT,
            target=FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )

        # select targets:
        st_sel = (x > 0) & (ct > 0.0)
        if np.any(st_sel):
            # get ws:
            ws = super().get_data(
                FV.REWS,
                target=FC.STATE_TARGET,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )

            # get TI:
            ti = super().get_data(
                FV.TI,
                target=FC.STATE_TARGET,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )

            # get alpha:
            alpha = super().get_data(
                FV.PA_ALPHA,
                target=FC.STATE_TARGET,
                lookup="ws",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )

            # get beta:
            beta = super().get_data(
                FV.PA_BETA,
                target=FC.STATE_TARGET,
                lookup="ws",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )

            # apply filter:
            x = x[st_sel]
            D = D[st_sel]
            ct = ct[st_sel]
            ws = ws[st_sel]
            ti = ti[st_sel]
            k = k[st_sel]
            gamma = gamma[st_sel]
            alpha = alpha[st_sel]
            beta = beta[st_sel]

            # calc theta_c0, Eq. (6.12):
            cosg = np.cos(gamma)
            twoac = 2 * self.induction.ct2a(ct * cosg)
            theta = 0.3 * gamma / cosg * twoac

            # calculate x0, Eq. (7.3):
            twoa = 2 * self.induction.ct2a(ct)
            x0 = D * cosg * (2 - twoa) / (np.sqrt(2) * (4 * alpha * ti + beta * twoa))
            out[self.X0] = x0

            # calcuate sigma, Eq. (7.2):
            sigma_y0 = D * cosg / np.sqrt(8)
            simga_z0 = D / np.sqrt(8)
            sigma_y = k * (x - x0) + sigma_y0
            sigma_z = k * (x - x0) + simga_z0

            # calc near wake data:
            near = x < x0
            out[self.NEAR] = near
            if np.any(near):
                # apply filter:
                ctn = ct[near]
                cosgn = cosg[near]
                twoan = twoa[near]
                twoacn = twoac[near]

                # initial velocity deficits, Eq. (6.4):
                uR = 0.5 * ctn * cosgn / twoacn

                # constant potential core value, Eq. (6.7):
                u0 = 1 - twoan

                # compute potential core shape, for later, Eq. (6.13):
                d = x[near] / x0[near]
                r_pc_0 = 0.5 * D[near] * np.sqrt(uR / u0)  # radius at x=0
                r_pc = r_pc_0 - d * r_pc_0  # potential core radius

                # memorize near wake data:
                out[self.R_PC] = r_pc
                out[self.R_PC_S] = d * sigma_y0[near]
                out[self.DELTA_NEAR] = theta[near] * x[near]
                out[self.AMPL_NEAR] = u0 - 1

                # cleanup:
                del ctn, cosgn, uR, u0, d, r_pc_0, r_pc, twoan, twoacn

            # calc far wake data:
            far = ~near
            if np.any(far):
                # apply filter:
                ws = ws[far]
                ct = ct[far]
                sigma_y = sigma_y[far]
                sigma_z = sigma_z[far]
                cosg = cosg[far]
                D = D[far]
                theta = theta[far]
                x0 = x0[far]
                k = k[far]
                twoa = twoa[far]

                # calculate delta, Eq. (7.4):
                sqct = np.sqrt(ct)
                sqsd = np.sqrt(8 * sigma_y * sigma_z / (cosg * D**2))
                delta = theta * x0 + (
                    D
                    * theta
                    / 14.7
                    * np.sqrt(cosg / (ct * k**2))
                    * (2.9 + 1.3 * (1 - twoa) - ct)
                    * np.log(
                        ((1.6 + sqct) * (1.6 * sqsd - sqct))
                        / ((1.6 - sqct) * (1.6 * sqsd + sqct))
                    )
                )

                # calculate amplitude, Eq. (7.1):
                ct_eff = ct * cosg * D**2 / (8 * sigma_y * sigma_z)
                ampl = np.maximum(-2 * self.induction.ct2a(ct_eff), -1)

                # memorize far wake data:
                out[self.AMPL_FAR] = ampl
                out[self.DELTA_FAR] = delta
                out[self.SIGMA_Y_FAR] = sigma_y
                out[self.SIGMA_Z_FAR] = sigma_z

        # update mdata:
        out[self.ST_SEL] = st_sel
        mdata[self.MDATA_KEY] = out

    def has_data(self, mdata, downwind_index, x):
        """
        Check if data exists

        Parameters
        ----------
        mdata: foxes.core.Data
            The model data
        downwind_index: numpy.ndarray
            For each state, one turbine index for the
            wake causing turbine. Shape: (n_states,)
        x: numpy.ndarray
            The x values, shape: (n_states, n_points)

        Returns
        -------
        check: bool
            True if data exists

        """
        check = (
            mdata[FC.STATE][0],
            downwind_index,
            x.shape,
        )
        return self.MDATA_KEY in mdata and mdata[self.MDATA_KEY][self.CHECK] == check

    def get_data(self, key, mdata):
        """
        Return data entry

        Parameters
        ----------
        key: str
            The data key
        mdata: foxes.core.Data
            The model data

        Returns
        -------
        data: numpy.ndarray
            The data

        """
        return mdata[self.MDATA_KEY][key]

    def clean(self, mdata):
        """
        Clean all data
        """
        del mdata[self.MDATA_KEY]


class Bastankhah2016(DistSlicedWakeModel):
    """
    The Bastankhah 2016 wake model

    Notes
    -----
    Reference:
    "Experimental and theoretical study of wind turbine wakes in yawed conditions"
    Majid Bastankhah, Fernando Porté-Agel
    https://doi.org/10.1017/jfm.2016.595

    Attributes
    ----------
    model: Bastankhah2016Model
        The model for computing common data
    model_pars: dict
        Model parameters
    YAWM: float
        The yaw misalignment YAWM. If not given here
        it will be searched in the farm data.
    alpha: float
        model parameter used to determine onset of far wake region
    beta: float
        model parameter used to determine onset of far wake region
    induction: foxes.core.AxialInductionModel or str
        The induction model
    wake_k: dict, optional
        Parameters for the WakeK class

    :group: models.wake_models.wind

    """

    def __init__(
        self,
        superposition,
        alpha=0.58,
        beta=0.07,
        induction="Madsen",
        **wake_k,
    ):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The wind deficit superposition
        ct_max: float
            The maximal value for ct, values beyond will be limited
            to this number, by default 0.9999
        alpha: float
            model parameter used to determine onset of far wake region
        beta: float
            model parameter used to determine onset of far wake region
        induction: foxes.core.AxialInductionModel or str
            The induction model
        wake_k: dict, optional
            Parameters for the WakeK class

        """
        super().__init__(superpositions={FV.WS: superposition})

        self.model = None
        self.alpha = alpha
        self.beta = beta
        self.induction = induction
        self.wake_k = WakeK(**wake_k)

        setattr(self, FV.YAWM, 0.0)

    def __repr__(self):
        iname = self.induction
        s = f"{type(self).__name__}"
        s += f"({self.superpositions[FV.WS]}, induction={iname}, "
        s += self.wake_k.repr() + ")"
        return s

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return [self.wake_k, self.model]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if not self.initialized:
            self.model = Bastankhah2016Model(
                alpha=self.alpha, beta=self.beta, induction=self.induction
            )
        super().initialize(algo, verbosity, force)

    def calc_wakes_x_yz(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        yz,
    ):
        """
        Calculate wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        yz: numpy.ndarray
            The yz values for each x value, shape:
            (n_states, n_targets, n_yz_per_target, 2)

        Returns
        -------
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, shape: (n_st_sel, n_yz_per_target)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        # prepare:
        n_y_per_z = yz.shape[2]

        # calculate model data:
        if not self.model.has_data(mdata, downwind_index, x):
            # get gamma:
            gamma = self.get_data(
                FV.YAWM,
                FC.STATE_TARGET,
                lookup="ws",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                upcast=True,
                downwind_index=downwind_index,
            )
            gamma *= np.pi / 180

            # get k:
            k = self.wake_k(
                FC.STATE_TARGET,
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                upcast=True,
                downwind_index=downwind_index,
            )

            # run calculation:
            self.model.calc_data(algo, mdata, fdata, tdata, downwind_index, x, gamma, k)

        # select targets:
        st_sel = self.model.get_data(Bastankhah2016Model.ST_SEL, mdata)
        n_sp_sel = np.sum(st_sel)
        wdeltas = {FV.WS: np.zeros((n_sp_sel, n_y_per_z), dtype=FC.DTYPE)}
        if np.any(st_sel):
            # apply filter:
            yz = yz[st_sel]

            # collect data:
            near = self.model.get_data(Bastankhah2016Model.NEAR, mdata)
            far = ~near

            # near wake:
            if np.any(near):
                # collect data:
                ampl = self.model.get_data(Bastankhah2016Model.AMPL_NEAR, mdata)
                r_pc = self.model.get_data(Bastankhah2016Model.R_PC, mdata)
                s = self.model.get_data(Bastankhah2016Model.R_PC_S, mdata)

                # radial dependency:
                r = np.linalg.norm(yz[near], axis=-1)
                rfactor = np.ones_like(r)
                sel_oc = np.where(r > r_pc[:, None])
                r = r[sel_oc]
                r_pc = r_pc[sel_oc[0]]
                s = s[sel_oc[0]]
                rfactor[sel_oc] = np.exp(-0.5 * ((r - r_pc) / s) ** 2)

                # set deficit, Eq. (6.13):
                wdeltas[FV.WS][near] = ampl[:, None] * rfactor

            # far wake:
            if np.any(far):
                # apply filter:
                yz = yz[far]

                # collect data:
                ampl = self.model.get_data(Bastankhah2016Model.AMPL_FAR, mdata)[:, None]
                sigma_y = self.model.get_data(Bastankhah2016Model.SIGMA_Y_FAR, mdata)[
                    :, None
                ]
                sigma_z = self.model.get_data(Bastankhah2016Model.SIGMA_Z_FAR, mdata)[
                    :, None
                ]

                # set deficit, Eq. (7.1):
                y = yz[..., 0]
                z = yz[..., 1]
                wdeltas[FV.WS][far] = ampl * (
                    np.exp(-0.5 * (y / sigma_y) ** 2)
                    * np.exp(-0.5 * (z / sigma_z) ** 2)
                )

        return wdeltas, st_sel



import numpy as np
from foxes.core import PointDataModel, TData
import foxes.variables as FV
import foxes.constants as FC


class PointWakesCalculation(PointDataModel):
    """
    This model calculates wake effects at points of interest.

    Attributes
    ----------
    pvars: list of str
        The variables of interest
    emodels: foxes.core.PointDataModelList
        The extra evaluation models
    emodels_cpars: list of dict
        The calculation parameters for extra models
    wake_models: list of foxes.core.WakeModel
        The wake models to be used

    :group: algorithms.downwind.models

    """

    def __init__(self, emodels=None, emodels_cpars=None, wake_models=None):
        """
        Constructor.

        Parameters
        ----------
        emodels: foxes.core.PointDataModelList, optional
            The extra evaluation models
        emodels_cpars: list of dict, optional
            The calculation parameters for extra models
        wake_models: list of foxes.core.WakeModel, optional
            Specific wake models to be used

        """
        super().__init__()
        self.pvars = None
        self.emodels = emodels
        self.emodels_cpars = emodels_cpars
        self.wake_models = wake_models

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return [self.emodels] if self.emodels is not None else []

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(algo, verbosity)
        self.pvars = algo.states.output_point_vars(algo)

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.pvars

    def calculate(self, algo, mdata, fdata, tdata, downwind_index=None):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        tdata: foxes.core.Data
            The target point data
        downwind_index: int
            The index in the downwind order of the wake
            causing turbine

        Returns
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        res = {}
        wmodels = (
            algo.wake_models.values() if self.wake_models is None else self.wake_models
        )
        for wmodel in wmodels:
            pwake = algo.partial_wakes[wmodel.name]
            gmodel = algo.ground_models[wmodel.name]

            wdeltas = gmodel.new_point_wake_deltas(algo, mdata, fdata, tdata, wmodel)

            if len(set(self.pvars).intersection(wdeltas.keys())):

                if downwind_index is None:
                    for oi in range(fdata.n_turbines):
                        gmodel.contribute_to_point_wakes(
                            algo, mdata, fdata, tdata, oi, wdeltas, wmodel
                        )
                else:
                    gmodel.contribute_to_point_wakes(
                        algo, mdata, fdata, tdata, downwind_index, wdeltas, wmodel
                    )

                for v in self.pvars:
                    if v not in res and v in tdata:
                        res[v] = tdata[v].copy()

                gmodel.finalize_point_wakes(algo, mdata, fdata, res, wdeltas, wmodel)

                for v in res.keys():
                    if v in wdeltas:
                        res[v] += wdeltas[v]

        for v in res.keys():
            tdata[v] = res[v]

        if self.emodels is not None:
            self.emodels.calculate(algo, mdata, fdata, tdata, self.emodels_cpars)

        return {v: tdata[v] for v in self.pvars}



from foxes.core import WakeK
from foxes.models.wake_models.top_hat import TopHatWakeModel
import foxes.variables as FV
import foxes.constants as FC


class JensenWake(TopHatWakeModel):
    """
    The Jensen wake model.

    Attributes
    ----------
    wake_k: foxes.core.WakeK
        Handler for the wake growth parameter k

    :group: models.wake_models.wind

    """

    def __init__(self, superposition, induction="Betz", **wake_k):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The wind deficit superposition
        induction: foxes.core.AxialInductionModel or str
            The induction model
        wake_k: dict, optional
            Parameters for the WakeK class

        """
        super().__init__(superpositions={FV.WS: superposition}, induction=induction)
        self.wake_k = WakeK(**wake_k)

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        s = f"{type(self).__name__}"
        s += f"({self.superpositions[FV.WS]}, induction={iname}, "
        s += self.wake_k.repr() + ")"
        return s

    def calc_wake_radius(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        ct,
    ):
        """
        Calculate the wake radius, depending on x only (not r).

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        ct: numpy.ndarray
            The ct values of the wake-causing turbines,
            shape: (n_states, n_targets)

        Returns
        -------
        wake_r: numpy.ndarray
            The wake radii, shape: (n_states, n_targets)

        """
        D = self.get_data(
            FV.D,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=False,
        )

        k = self.wake_k(
            FC.STATE_TARGET,
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=False,
        )

        return D / 2 + k * x

    def calc_centreline(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        x,
        wake_r,
        ct,
    ):
        """
        Calculate centre line results of wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)
        x: numpy.ndarray
            The x values, shape: (n_st_sel,)
        wake_r: numpy.ndarray
            The wake radii, shape: (n_st_sel,)
        ct: numpy.ndarray
            The ct values of the wake-causing turbines,
            shape: (n_st_sel,)

        Returns
        -------
        cl_del: dict
            The centre line wake deltas. Key: variable name str,
            varlue: numpy.ndarray, shape: (n_st_sel,)

        """
        R = (
            self.get_data(
                FV.D,
                FC.STATE_TARGET,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
            )[st_sel]
            / 2
        )

        twoa = 2 * self.induction.ct2a(ct)

        return {FV.WS: -((R / wake_r) ** 2) * twoa}



import numpy as np

from foxes.core import TurbineModel
import foxes.constants as FC
import foxes.variables as FV


class SetFarmVars(TurbineModel):
    """
    Set farm data variables to given data.

    Attributes
    ----------
    vars: list of str
        The variables to be set

    :group: models.turbine_models

    """

    def __init__(self, pre_rotor=False):
        """
        Constructor.

        Parameters
        ----------
        pre_rotor: bool
            Flag for running this model before
            running the rotor model.

        """
        super().__init__(pre_rotor=pre_rotor)
        self.reset()

    def add_var(self, var, data):
        """
        Add data for a variable.

        Parameters
        ----------
        var: str
            The variable name
        data: numpy.ndarray
            The data, shape: (n_states, n_turbines)

        """
        self.vars.append(var)
        self._vdata.append(np.asarray(data, dtype=FC.DTYPE))

    def reset(self):
        """
        Remove all variables.
        """
        self.vars = []
        self._vdata = []

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.vars

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        idata = super().load_data(algo, verbosity)

        for i, v in enumerate(self.vars):
            data = np.full((algo.n_states, algo.n_turbines), np.nan, dtype=FC.DTYPE)
            vdata = self._vdata[i]

            # handle special case of call during vectorized optimization:
            if (
                np.ndim(vdata)
                and vdata.shape[0] != algo.n_states
                and hasattr(algo.states, "n_pop")
            ):
                n_pop = algo.states.n_pop
                n_ost = algo.states.states.size()
                n_trb = algo.n_turbines
                vdata = np.zeros((n_pop, n_ost, n_trb), dtype=FC.DTYPE)
                vdata[:] = self._vdata[i][None, :]
                vdata = vdata.reshape(n_pop * n_ost, n_trb)

            data[:] = vdata
            idata["data_vars"][self.var(v)] = ((FC.STATE, FC.TURBINE), data)

        return idata

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        if self.pre_rotor:
            order = np.s_[:]
            ssel = np.s_[:]
        else:
            order = fdata[FV.ORDER]
            ssel = fdata[FV.ORDER_SSEL]

        bsel = np.zeros((fdata.n_states, fdata.n_turbines), dtype=bool)
        bsel[st_sel] = True

        for v in self.vars:
            data = mdata[self.var(v)][ssel, order]
            hsel = ~np.isnan(data)
            tsel = bsel & hsel

            # special case of turbine positions:
            if v in [FV.X, FV.Y]:
                i = [FV.X, FV.Y].index(v)
                for ti in np.where(tsel)[1]:
                    t = algo.farm.turbines[ti]
                    if len(t.xy.shape) == 1:
                        xy = np.zeros((algo.n_states, 2), dtype=FC.DTYPE)
                        xy[:] = t.xy[None, :]
                        t.xy = xy
                    i0 = fdata.states_i0()
                    hsel = tsel[:, ti]
                    ssel = i0 + np.where(hsel)[0]
                    t.xy[ssel, i] = data[hsel, ti]

            # special case of rotor diameter and hub height:
            if v in [FV.D, FV.H]:
                for ti in np.where(tsel)[1]:
                    t = algo.farm.turbines[ti]
                    x = np.zeros(algo.n_states, dtype=FC.DTYPE)
                    if v == FV.D:
                        x[:] = t.D
                        t.D = x
                    else:
                        x[:] = t.H
                        t.H = x
                    i0 = fdata.states_i0()
                    hsel = tsel[:, ti]
                    ssel = i0 + np.where(hsel)[0]
                    x[ssel] = data[hsel, ti]

            fdata[v][tsel] = data[tsel]

        return {v: fdata[v] for v in self.vars}



"""
Model collection.
"""

from . import turbine_types
from . import rotor_models
from . import turbine_models
from . import farm_models
from . import partial_wakes
from . import wake_frames
from . import wake_models
from . import wake_superpositions
from . import farm_controllers
from . import vertical_profiles
from . import point_models
from . import axial_induction
from . import ground_models

from .model_book import ModelBook



"""
Abstract classes and core functionality.
"""

from .data import Data, MData, FData, TData
from .model import Model
from .data_calc_model import DataCalcModel
from .states import States, ExtendedStates
from .wind_farm import WindFarm
from .algorithm import Algorithm
from .farm_data_model import FarmDataModel, FarmDataModelList
from .point_data_model import PointDataModel, PointDataModelList
from .rotor_model import RotorModel
from .farm_model import FarmModel
from .turbine_model import TurbineModel
from .turbine_type import TurbineType
from .farm_controller import FarmController
from .turbine import Turbine
from .partial_wakes_model import PartialWakesModel
from .wake_frame import WakeFrame
from .wake_model import WakeModel, TurbineInductionModel, WakeK
from .wake_superposition import WakeSuperposition
from .vertical_profile import VerticalProfile
from .axial_induction_model import AxialInductionModel
from .ground_model import GroundModel



import json
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

import foxes.constants as FC
import foxes.variables as FV
from foxes.output.output import Output


class FarmLayoutOutput(Output):
    """
    Plot the farm layout

    Attributes
    ----------
    farm: foxes.WindFarm
        The wind farm
    fres: xarray.Dataset
        The wind farm calculation results
    from_res: bool
        Flag for coordinates from results data
    results_state: int
        The state index, for from_res
    D: float
        The rotor diameter, if not from data

    :group: output

    """

    def __init__(
        self, farm, farm_results=None, from_results=False, results_state=None, D=None
    ):
        """
        Constructor.

        Parameters
        ----------
        farm: foxes.WindFarm
            The wind farm
        farm_results: xarray.Dataset, optional
            The wind farm calculation results
        from_results: bool, optional
            Flag for coordinates from results data
        results_state: int, optional
            The state index, for from_res
        D: float, optional
            The rotor diameter, if not from data

        """
        self.farm = farm
        self.fres = farm_results
        self.from_res = from_results
        self.rstate = results_state
        self.D = D

        if from_results and farm_results is None:
            raise ValueError(f"Missing farm_results for switch from_results.")

        if from_results and results_state is None:
            raise ValueError(f"Please specify results_state for switch from_results.")

    def get_layout_data(self):
        """
        Returns wind farm layout.

        Returns
        -------
        numpy.ndarray :
            The wind farm layout, shape:
            (n_turbines, 3) where the 3
            represents x, y, h

        """

        data = np.zeros([self.farm.n_turbines, 3], dtype=FC.DTYPE)

        if self.from_res:
            data[:, 0] = self.fres[FV.X][self.rstate]
            data[:, 1] = self.fres[FV.Y][self.rstate]
            data[:, 2] = self.fres[FV.H][self.rstate]

        else:
            for ti, t in enumerate(self.farm.turbines):
                data[ti, :2] = t.xy
                data[ti, 2] = t.H

        return data

    def get_layout_dict(self):
        """
        Returns wind farm layout.

        Returns
        -------
        dict :
            The wind farm layout in dict
            format, as in json output

        """

        data = self.get_layout_data()

        out = {self.farm.name: {}}
        for ti, p in enumerate(data):
            t = self.farm.turbines[ti]
            out[self.farm.name][t.name] = {
                "id": t.index,
                "name": t.name,
                "UTMX": p[0],
                "UTMY": p[1],
            }

        return out

    def get_figure(
        self,
        color_by=None,
        fontsize=8,
        figsize=None,
        annotate=1,
        title=None,
        fig=None,
        ax=None,
        normalize_D=False,
        ret_im=False,
        bargs={},
        anno_delx=0,
        anno_dely=0,
        **kwargs,
    ):
        """
        Creates farm layout figure.

        Parameters
        ----------
        color_by: str, optional
            Set scatter color by variable results.
            Use "mean_REWS" etc for means, also
            min, max, sum. All wrt states
        fontsize: int, optional
            Size of the turbine numbers
        figsize: tuple, optional
            The figsize for plt.Figure
        annotate: int, optional
            Turbine index printing, Choices:
            0 = No annotation
            1 = Turbine indices
            2 = Turbine names
        title: str, optional
            The plot title, or None for automatic
        fig: matplotlib.pyplot.Figure, optional
            The figure object to which to add
        ax: matplotlib.pyplot.Axis, optional
            The axis object, to which to add
        normalize_D: bool
            Normalize x, y wrt rotor diameter
        ret_im: bool
            Flag for returned image object
        bargs: dict
            Arguments for boundary plotting
        anno_delx: float
            The annotation delta x
        anno_dely: float
            The annotation delta y
        kwargs: dict, optional
            Parameters forwarded to `matplotlib.pyplot.scatter`

        Returns
        -------
        ax: matplotlib.pyplot.Axis
            The axis object
        im: matplotlib.pyplot.PathCollection, optional
            The image object

        """

        if fig is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
        else:
            ax = fig.axes[0] if ax is None else ax

        D = self.D
        x = None
        if self.farm.n_turbines:
            if normalize_D and D is None:
                if self.from_res:
                    if self.fres[FV.D].min() != self.fres[FV.D].max():
                        raise ValueError(
                            f"Expecting uniform D, found {self.fres[FV.D]}"
                        )
                    D = self.fres[FV.D][0]
                else:
                    D = None
                    for ti, t in enumerate(self.farm.turbines):
                        hD = t.D
                        if D is None:
                            D = hD
                        elif D != hD:
                            raise ValueError(
                                f"Turbine {ti} has wrong rotor diameter, expecting D = {D} m, found D = {hD} m"
                            )
                    if D is None:
                        raise ValueError(
                            f"Variable '{FV.D}' not found in turbines. Maybe set explicitely, or try from_results?"
                        )

            data = self.get_layout_data()
            x = data[:, 0] / D if normalize_D else data[:, 0]
            y = data[:, 1] / D if normalize_D else data[:, 1]
            n = range(len(x))

            kw = {"c": "orange"}
            kw.update(**kwargs)

            if color_by is not None:
                if self.fres is None:
                    raise ValueError(f"Missing farm_results for color_by '{color_by}'")
                if color_by[:5] == "mean_":
                    kw["c"] = np.einsum(
                        "st,st->t", self.fres[color_by[5:]], self.fres[FV.WEIGHT]
                    )
                elif color_by[:4] == "sum_":
                    kw["c"] = np.sum(self.fres[color_by[4:]], axis=0)
                elif color_by[:4] == "min_":
                    kw["c"] = np.min(self.fres[color_by[4:]], axis=0)
                elif color_by[:4] == "max_":
                    kw["c"] = np.max(self.fres[color_by[4:]], axis=0)
                else:
                    raise KeyError(
                        f"Unknown color_by '{color_by}'. Choose: mean_X, sum_X, min_X, max_X, where X is a farm_results variable"
                    )

            im = ax.scatter(x, y, **kw)

            if annotate == 1:
                for i, txt in enumerate(n):
                    ax.annotate(
                        int(txt), (x[i] + anno_delx, y[i] + anno_dely), size=fontsize
                    )
            elif annotate == 2:
                for i, t in enumerate(self.farm.turbines):
                    ax.annotate(
                        t.name, (x[i] + anno_delx, y[i] + anno_dely), size=fontsize
                    )

        if self.farm.boundary is not None:
            hbargs = {"fill_mode": "inside_lightgray"}
            hbargs.update(bargs)
            self.farm.boundary.add_to_figure(ax, **hbargs)

        ti = (
            title
            if title is not None
            else (
                self.farm.name
                if D is None or not normalize_D
                else f"{self.farm.name} (D = {D} m)"
            )
        )
        ax.set_title(ti)

        ax.set_xlabel("x [m]" if not normalize_D else "x [D]")
        ax.set_ylabel("y [m]" if not normalize_D else "y [D]")
        ax.grid()

        # if len(self.farm.boundary_geometry) \
        #    or ( min(x) != max(x) and min(y) != max(y) ):
        if x is None or (min(x) != max(x) and min(y) != max(y)):
            ax.set_aspect("equal", adjustable="box")

        ax.autoscale_view(tight=True)

        if color_by is not None:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            fig.colorbar(im, cax=cax)

        if ret_im:
            return ax, im

        return ax

    def write_plot(self, file_path=None, fontsize=8, **kwargs):
        """
        Writes the layout plot to file.

        The kwargs are forwarded to self.get_figure

        Parameters
        ----------
        file_path: str
            The file into which to plot, or None
            for default
        fontsize: int
            Size of the turbine numbers

        """

        ax = self.get_figure(fontsize=fontsize, ret_im=False, **kwargs)
        fig = ax.get_figure()

        fname = file_path if file_path is not None else self.farm.name + ".png"
        fig.savefig(fname, bbox_inches="tight")

        plt.close(fig)

    def write_xyh(self, file_path=None):
        """
        Writes xyh layout file.

        Parameters
        ----------
        file_path: str
            The file into which to plot, or None
            for default

        """

        data = self.get_layout_data()

        fname = file_path if file_path is not None else self.farm.name + ".xyh"
        np.savetxt(fname, data, header="x y h")

    def write_csv(self, file_path=None):
        """
        Writes csv layout file.

        Parameters
        ----------
        file_path: str
            The file into which to plot, or None
            for default

        """

        data = self.get_layout_data()

        fname = file_path if file_path is not None else self.farm.name + ".csv"

        lyt = pd.DataFrame(
            index=range(len(data)), columns=["id", "name", "x", "y", "h", "D"]
        )
        lyt.index.name = "index"
        lyt["id"] = [t.info["id"] for t in self.farm.turbines]
        lyt["name"] = [t.info["name"] for t in self.farm.turbines]
        lyt["x"] = data[:, 0]
        lyt["y"] = data[:, 1]
        lyt["h"] = data[:, 2]
        lyt["D"] = [t.D for t in self.farm.turbines]

        lyt.to_csv(fname)

    def write_json(self, file_path=None):
        """
        Writes xyh layout file.

        Parameters
        ----------
        file_path: str
            The file into which to plot, or None
            for default

        """

        data = self.get_layout_dict()

        fname = file_path if file_path is not None else self.farm.name + ".json"
        with open(fname, "w") as outfile:
            json.dump(data, outfile, indent=4)



import numpy as np

from foxes.core import WakeSuperposition
import foxes.variables as FV
import foxes.constants as FC


class WSProduct(WakeSuperposition):
    """
    Product supersposition of wind deficit results

    This is based on the idea that the dimensionless
    wind deficit should be rescaled with the wake
    corrected wind field, rather than the rotor
    equivalent wind speed.

    Source: https://arxiv.org/pdf/2010.03873.pdf
            Equation (8)

    Attributes
    ----------
    lim_low: float
        Lower limit of the final waked wind speed
    lim_high: float
        Upper limit of the final waked wind speed

    :group: models.wake_superpositions

    """

    def __init__(self, lim_low=None, lim_high=None):
        """
        Constructor.

        Parameters
        ----------
        lim_low: float
            Lower limit of the final waked wind speed
        lim_high: float
            Upper limit of the final waked wind speed

        """
        super().__init__()

        self.lim_low = lim_low
        self.lim_high = lim_high

    def __repr__(self):
        a = f"lim_low={self.lim_low}, lim_high={self.lim_high}"
        return f"{type(self).__name__}({a})"

    def input_farm_vars(self, algo):
        """
        The variables which are needed for running
        the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        input_vars: list of str
            The input variable names

        """
        return [FV.AMB_REWS] if self.scale_amb else [FV.REWS]

    def add_wake(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        variable,
        wake_delta,
        wake_model_result,
    ):
        """
        Add a wake delta to previous wake deltas,
        at rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        st_sel: numpy.ndarray of bool
            The selection of targets, shape: (n_states, n_targets)
        variable: str
            The variable name for which the wake deltas applies
        wake_delta: numpy.ndarray
            The original wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)
        wake_model_result: numpy.ndarray
            The new wake deltas of the selected rotors,
            shape: (n_st_sel, n_tpoints, ...)

        Returns
        -------
        wdelta: numpy.ndarray
            The updated wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)

        """
        if variable not in [FV.REWS, FV.REWS2, FV.REWS3, FV.WS]:
            raise ValueError(
                f"Superposition '{self.name}': Expecting wind speed variable, got {variable}"
            )

        if np.any(st_sel):
            if np.max(np.abs(wake_delta)) < 1e-14:
                wake_delta[:] = 1

            wake_delta[st_sel] *= 1 + wake_model_result

        return wake_delta

    def calc_final_wake_delta(
        self,
        algo,
        mdata,
        fdata,
        variable,
        amb_results,
        wake_delta,
    ):
        """
        Calculate the final wake delta after adding all
        contributions.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        variable: str
            The variable name for which the wake deltas applies
        amb_results: numpy.ndarray
            The ambient results at targets,
            shape: (n_states, n_targets, n_tpoints)
        wake_delta: numpy.ndarray
            The wake deltas at targets, shape:
            (n_states, n_targets, n_tpoints)

        Returns
        -------
        final_wake_delta: numpy.ndarray
            The final wake delta, which will be added to the ambient
            results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)

        """
        w = amb_results * (wake_delta - 1)
        if self.lim_low is not None:
            w = np.maximum(w, self.lim_low - amb_results)
        if self.lim_high is not None:
            w = np.minimum(w, self.lim_high - amb_results)
        return w



import numpy as np

from foxes.core import States
import foxes.variables as FV
import foxes.constants as FC


class ScanWS(States):
    """
    A given list of wind speeds, all other variables are fixed.

    Parameters
    ----------
    wd: float
        The wind direction
    ti: float
        The TI value
    rho: float
        The air density

    :group: input.states

    """

    def __init__(self, ws_list, wd, ti=None, rho=None):
        """
        Constructor.

        Parameters
        ----------
        ws_list: array_like
            The wind speed values
        wd: float
            The wind direction
        ti: float, optional
            The TI value
        rho: float, optional
            The air density

        """
        super().__init__()

        self._wsl = np.array(ws_list)
        self.N = len(ws_list)
        self.wd = wd
        self.ti = ti
        self.rho = rho

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        self.WS = self.var(FV.WS)

        idata = super().load_data(algo, verbosity)
        idata["data_vars"][self.WS] = ((FC.STATE,), self._wsl)

        return idata

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self.N

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        pvars = [FV.WS]
        if self.wd is not None:
            pvars.append(FV.WD)
        if self.ti is not None:
            pvars.append(FV.TI)
        if self.rho is not None:
            pvars.append(FV.RHO)
        return pvars

    def weights(self, algo):
        """
        The statistical weights of all states.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        weights: numpy.ndarray
            The weights, shape: (n_states, n_turbines)

        """
        return np.full((self.N, algo.n_turbines), 1.0 / self.N, dtype=FC.DTYPE)

    def calculate(self, algo, mdata, fdata, tdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        tdata[FV.WS] = np.zeros_like(tdata[FC.TARGETS][..., 0])
        tdata[FV.WS][:] = mdata[self.WS][:, None, None]

        if self.wd is not None:
            tdata[FV.WD] = np.full_like(tdata[FV.WS], self.wd)
        if self.ti is not None:
            tdata[FV.TI] = np.full_like(tdata[FV.WS], self.ti)
        if self.rho is not None:
            tdata[FV.RHO] = np.full_like(tdata[FV.WS], self.rho)

        return {v: tdata[v] for v in self.output_point_vars(algo)}



import numpy as np
from abc import abstractmethod

from .data_calc_model import DataCalcModel
import foxes.constants as FC


class PointDataModel(DataCalcModel):
    """
    Abstract base class for models that modify
    point based data.

    :group: core

    """

    @abstractmethod
    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return []

    def ensure_variables(self, algo, mdata, fdata, tdata):
        """
        Add variables to tdata, initialized with NaN

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        tdata: foxes.core.Data
            The target point data

        """
        for v in self.output_point_vars(algo):
            if v not in tdata:
                tdata[v] = np.full(
                    (tdata.n_states, tdata.n_targets, tdata.n_tpoints),
                    np.nan,
                    dtype=FC.DTYPE,
                )
                tdata.dims[v] = (FC.STATE, FC.TARGET, FC.TPOINT)

    @abstractmethod
    def calculate(self, algo, mdata, fdata, tdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        pass

    def run_calculation(self, algo, *data, out_vars, **calc_pars):
        """
        Starts the model calculation in parallel, via
        xarray's `apply_ufunc`.

        Typically this function is called by algorithms.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        *data: tuple of xarray.Dataset
            The input data
        out_vars: list of str
            The calculation output variables
        **calc_pars: dict, optional
            Additional arguments for the `calculate` function

        Returns
        -------
        results: xarray.Dataset
            The calculation results

        """
        results = super().run_calculation(
            algo,
            *data,
            out_vars=out_vars,
            loop_dims=[FC.STATE, FC.TARGET],
            out_core_vars=[FC.TPOINT, FC.VARS],
            **calc_pars,
        )
        if results.sizes[FC.TPOINT] != 1:
            raise ValueError(
                f"PointDataModel '{self.name}': Expecting dimension '{FC.TPOINT}' of size 1, found {results.sizes[FC.TPOINT]}"
            )
        return results.sel({FC.TPOINT: 0}).rename({FC.TARGET: FC.POINT})

    def __add__(self, m):
        if isinstance(m, list):
            return PointDataModelList([self] + m)
        elif isinstance(m, PointDataModelList):
            return PointDataModelList([self] + m.models)
        else:
            return PointDataModelList([self, m])


class PointDataModelList(PointDataModel):
    """
    A list of point data models.

    By using the PointDataModelList the models'
    `calculate` functions are called together
    under one common call of xarray's `apply_ufunc`.

    Attributes
    ----------
    models: list of foxes.core.PointDataModel
        The model list

    :group: core

    """

    def __init__(self, models=[]):
        """
        Constructor.

        Parameters
        ----------
        models: list of foxes.core.PointDataModel
            The model list

        """
        super().__init__()
        self.models = models

    def append(self, model):
        """
        Add a model to the list

        Parameters
        ----------
        model: foxes.core.PointDataModel
            The model to add

        """
        self.models.append(model)

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return self.models

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        ovars = []
        for m in self.models:
            ovars += m.output_point_vars(algo)
        return list(dict.fromkeys(ovars))

    def calculate(self, algo, mdata, fdata, tdata, parameters=None):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        parameters: list of dict, optional
            A list of parameter dicts, one for each model

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        if parameters is None:
            parameters = [{}] * len(self.models)
        elif not isinstance(parameters, list):
            raise ValueError(
                f"{self.name}: Wrong parameters type, expecting list, got {type(parameters).__name__}"
            )
        elif len(parameters) != len(self.models):
            raise ValueError(
                f"{self.name}: Wrong parameters length, expecting list with {len(self.models)} entries, got {len(parameters)}"
            )

        for mi, m in enumerate(self.models):
            res = m.calculate(algo, mdata, fdata, tdata, **parameters[mi])
            tdata.update(res)

        return {v: tdata[v] for v in self.output_point_vars(algo)}



import numpy as np
import pandas as pd
from scipy.interpolate import interpn

from foxes.core import TurbineModel
from foxes.utils import PandasFileHelper
import foxes.constants as FC


class TableFactors(TurbineModel):
    """
    Multiplies variables by factors from a
    two dimensional table.

    The column names are expected to be numbers
    that represent the col_var variable.

    Attributes
    ----------
    data_source: str or pandas.DataFrame
        Either path to a file or data
    row_var: str
        The row-wise variable
    col_var: str
        The column-wise variable
    ovars: list of str
        The variables onto which the factors
        are multiplied

    :group: models.turbine_models

    """

    def __init__(
        self,
        data_source,
        row_var,
        col_var,
        output_vars,
        pd_file_read_pars={},
        **ipars,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame
            Either path to a file or data
        row_var: str
            The row-wise variable
        col_var: str
            The column-wise variable
        output_vars: list of str
            The variables onto which the factors
            are multiplied
        pd_file_read_pars: dict
            Parameters for pandas file reading
        ipars: dict, optional
            Parameters for scipy.interpolate.interpn

        """
        super().__init__()

        self.data_source = data_source
        self.row_var = row_var
        self.col_var = col_var
        self.ovars = output_vars
        self._rpars = pd_file_read_pars
        self._ipars = ipars

        self._cvals = None
        self._data = None

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.ovars

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(algo, verbosity)

        if isinstance(self.data_source, pd.DataFrame):
            self._data = self.data_source
        else:
            if verbosity > 0:
                print(f"{self.name}: Reading file {self.data_source}")
            rpars = dict(index_col=0)
            rpars.update(self._rpars)
            self._data = PandasFileHelper.read_file(self.data_source, **rpars)

        self._rvals = self._data.index.to_numpy(FC.DTYPE)
        self._cvals = self._data.columns.to_numpy(FC.DTYPE)
        self._data = self._data.to_numpy(FC.DTYPE)

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        n_sel = np.sum(st_sel)
        qts = np.zeros((n_sel, 2), dtype=FC.DTYPE)
        qts[:, 0] = fdata[self.row_var][st_sel]
        qts[:, 1] = fdata[self.col_var][st_sel]

        try:
            factors = interpn(
                (self._rvals, self._cvals), self._data, qts, **self._ipars
            )
        except ValueError as e:
            print(f"\nDATA       : ({self.row_var}, {self.col_var})")
            print(
                f"DATA BOUNDS: ({np.min(self._rvals)}, {np.min(self._cvals)}) -- ({np.max(self._rvals)}, {np.max(self._cvals)})"
            )
            print(
                f"VALUE BOUNDS: ({np.min(qts[:, 0]):.4f}, {np.min(qts[:, 1]):.4f}) -- ({np.max(qts[:, 0]):.4f}, {np.max(qts[:, 1]):.4f})\n"
            )
            raise e

        for v in self.output_farm_vars(algo):
            fdata[v][st_sel] *= factors

        return {v: fdata[v] for v in self.output_farm_vars(algo)}



import numpy as np

from foxes.output import SliceData
import foxes.variables as FV

from .get_fig import get_fig


class FlowPlots2D(SliceData):
    """
    Class for horizontal or vertical 2D flow plots

    :group: output.flow_plots_2d

    """

    def get_mean_fig_xy(
        self,
        var,
        xlabel="x [m]",
        ylabel="y [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        ret_data=False,
        animated=False,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a horizontal xy-plane.

        Parameters
        ----------
        var: str
            The variable name
        xlabel: str, optional
            The x axis label
        ylabel: str, optional
            The y axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at each `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        ret_data: bool
            Flag for returning image data
        animated: bool
            Switch for usage for an animation
        kwargs: dict, optional
            Additional parameters for SliceData.get_mean_data_xy

        Returns
        -------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet
        data: numpy.ndarray, optional
            The image data, shape: (n_x, n_y)

        """
        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        data, gdata = self.get_mean_data_xy(
            variables=variables,
            vmin={var: vmin} if vmin is not None else {},
            vmax={var: vmax} if vmax is not None else {},
            data_format="numpy",
            ret_grid=True,
            **kwargs,
        )
        x_pos, y_pos, z_pos, __ = gdata

        if title is None:
            title = f"States mean, z =  {int(np.round(z_pos))} m"

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[None, :, :, wdi],
                data[None, :, :, wsi],
            )
        )

        # create plot:
        out = get_fig(
            var=var,
            fig=fig,
            figsize=figsize,
            ax=ax,
            data=data[None, :, :, vi],
            si=0,
            s=None,
            levels=levels,
            x_pos=x_pos,
            y_pos=y_pos,
            cmap=cmap,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
            add_bar=add_bar,
            vlabel=vlabel,
            quiv=quiv,
            vmin=vmin,
            vmax=vmax,
            ret_state=ret_state,
            ret_im=ret_im,
            animated=animated,
        )

        if ret_data:
            out = list(out) if isinstance(out, tuple) else [out]
            return tuple(out + [data[:, :, 0]])

        return out

    def get_mean_fig_xz(
        self,
        var,
        x_direction=270,
        xlabel="x [m]",
        zlabel="z [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        ret_data=False,
        animated=False,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a horizontal xz-plane.

        Parameters
        ----------
        var: str
            The variable name
        x_direction: float, optional
            The direction of the x axis, 0 = north
        xlabel: str, optional
            The x axis label
        zlabel: str, optional
            The z axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at each `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        ret_data: bool
            Flag for returning image data
        animated: bool
            Switch for usage for an animation
        kwargs: dict, optional
            Additional parameters for SliceData.get_mean_data_xz

        Returns
        -------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet
        data: numpy.ndarray, optional
            The image data, shape: (n_x, n_y)

        """
        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        data, gdata = self.get_mean_data_xz(
            variables=variables,
            vmin={var: vmin} if vmin is not None else {},
            vmax={var: vmax} if vmax is not None else {},
            x_direction=x_direction,
            data_format="numpy",
            ret_grid=True,
            **kwargs,
        )
        x_pos, y_pos, z_pos, __ = gdata

        if title is None:
            title = f"States mean, x direction {x_direction}°, y =  {int(np.round(y_pos))} m"

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[None, :, :, wdi],
                data[None, :, :, wsi],
            )
        )

        # create plot:
        out = get_fig(
            var=var,
            fig=fig,
            figsize=figsize,
            ax=ax,
            data=data[None, :, :, vi],
            si=0,
            s=None,
            levels=levels,
            x_pos=x_pos,
            y_pos=z_pos,
            cmap=cmap,
            xlabel=xlabel,
            ylabel=zlabel,
            title=title,
            add_bar=add_bar,
            vlabel=vlabel,
            vmin=vmin,
            vmax=vmax,
            ret_state=ret_state,
            ret_im=ret_im,
            quiv=quiv,
            animated=animated,
        )

        if ret_data:
            out = list(out) if isinstance(out, tuple) else [out]
            return tuple(out + [data[:, :, 0]])

        return out

    def get_mean_fig_yz(
        self,
        var,
        x_direction=270,
        ylabel="x [m]",
        zlabel="z [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        ret_data=False,
        animated=False,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a horizontal yz-plane.

        Parameters
        ----------
        var: str
            The variable name
        x_direction: float, optional
            The direction of the x axis, 0 = north
        ylabel: str, optional
            The y axis label
        zlabel: str, optional
            The z axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at each `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        ret_data: bool
            Flag for returning image data
        animated: bool
            Switch for usage for an animation
        kwargs: dict, optional
            Additional parameters for SliceData.get_mean_data_yz

        Returns
        -------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet
        data: numpy.ndarray, optional
            The image data, shape: (n_x, n_y)

        """
        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        data, gdata = self.get_mean_data_yz(
            variables=variables,
            vmin={var: vmin} if vmin is not None else {},
            vmax={var: vmax} if vmax is not None else {},
            x_direction=x_direction,
            data_format="numpy",
            ret_grid=True,
            **kwargs,
        )
        x_pos, y_pos, z_pos, __ = gdata

        if title is None:
            title = f"States mean, x direction {x_direction}°, x =  {int(np.round(x_pos))} m"

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[None, :, :, wdi],
                data[None, :, :, wsi],
            )
        )

        # create plot:
        out = get_fig(
            var=var,
            fig=fig,
            figsize=figsize,
            ax=ax,
            data=data[None, :, :, vi],
            si=0,
            s=None,
            levels=levels,
            x_pos=y_pos,
            y_pos=z_pos,
            cmap=cmap,
            xlabel=ylabel,
            ylabel=zlabel,
            title=title,
            add_bar=add_bar,
            vlabel=vlabel,
            vmin=vmin,
            vmax=vmax,
            ret_state=ret_state,
            ret_im=ret_im,
            quiv=quiv,
            invert_axis="x",
            animated=animated,
        )

        if ret_data:
            out = list(out) if isinstance(out, tuple) else [out]
            return tuple(out + [data[:, :, 0]])

        return out

    def gen_states_fig_xy(
        self,
        var,
        xlabel="x [m]",
        ylabel="y [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        animated=False,
        rotor_color=None,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a horizontal xy-plane.

        Parameters
        ----------
        var: str
            The variable name
        xlabel: str, optional
            The x axis label
        ylabel: str, optional
            The y axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at each `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        animated: bool
            Switch for usage for an animation
        rotor_color: str, optional
            Indicate the rotor orientation by a colored line
        kwargs: dict, optional
            Additional parameters for SliceData.get_states_data_xy

        Yields
        ------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet

        """
        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        data, states, gdata = self.get_states_data_xy(
            variables=variables,
            vmin={var: vmin} if vmin is not None else {},
            vmax={var: vmax} if vmax is not None else {},
            data_format="numpy",
            ret_states=True,
            ret_grid=True,
            **kwargs,
        )
        x_pos, y_pos, z_pos, __ = gdata

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[..., wdi],
                data[..., wsi],
            )
        )

        # loop over states:
        for si, s in enumerate(states):
            if animated and si > 0 and vmin is not None and vmax is not None:
                add_bar = False
            if not animated and title is None:
                ttl = f"State {s}"
                ttl += f", z =  {int(np.round(z_pos))} m"
            elif callable(title):
                ttl = title(si, s)
            else:
                ttl = title

            # get data for show_turbines
            if rotor_color is not None:
                try:
                    turb_angle = self.fres[FV.AMB_WD][si] + self.fres[FV.YAWM][si]
                except KeyError:
                    turb_angle = self.fres[FV.AMB_WD][si]

                show_rotor_dict = {
                    "color": rotor_color,
                    "D": self.fres[FV.D][si],
                    "H": self.fres[FV.H][si],
                    "X": self.fres[FV.X][si],
                    "Y": self.fres[FV.Y][si],
                    "AMB_WD": self.fres[FV.AMB_WD][si],
                    "turb_angle": turb_angle,
                }
            else:
                show_rotor_dict = None

            out = get_fig(
                var=var,
                fig=fig,
                figsize=figsize,
                ax=ax,
                data=data[..., vi],
                si=si,
                s=s,
                levels=levels,
                x_pos=x_pos,
                y_pos=y_pos,
                cmap=cmap,
                xlabel=xlabel,
                ylabel=ylabel,
                title=ttl,
                add_bar=add_bar,
                vlabel=vlabel,
                vmin=vmin,
                vmax=vmax,
                quiv=quiv,
                ret_state=ret_state,
                ret_im=ret_im,
                animated=animated,
                show_rotor_dict=show_rotor_dict,
            )

            yield out

    def gen_states_fig_xz(
        self,
        var,
        x_direction=270.0,
        xlabel="x [m]",
        zlabel="z [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        animated=False,
        rotor_color=None,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a vertical xz-plane.

        Parameters
        ----------
        var: str
            The variable name
        x_direction: float, optional
            The direction of the x axis, 0 = north
        xlabel: str, optional
            The x axis label
        zlabel: str, optional
            The z axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at ech `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        animated: bool
            Switch for usage for an animation
        rotor_color: str, optional
            Indicate the rotor orientation by a colored line
        kwargs: dict, optional
            Additional parameters for SliceData.get_states_data_xz

        Yields
        ------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet

        """
        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        data, states, gdata = self.get_states_data_xz(
            variables=variables,
            vmin={var: vmin} if vmin is not None else {},
            vmax={var: vmax} if vmax is not None else {},
            data_format="numpy",
            ret_states=True,
            ret_grid=True,
            x_direction=x_direction,
            **kwargs,
        )
        x_pos, y_pos, z_pos, __ = gdata

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[..., wdi],
                data[..., wsi],
            )
        )

        # loop over states:
        for si, s in enumerate(states):
            if animated and si > 0 and vmin is not None and vmax is not None:
                add_bar = False
            if not animated and title is None:
                ttl = f"State {s}"
                ttl += f", x direction = {x_direction}°"
                ttl += f", y =  {int(np.round(y_pos))} m"
            elif callable(title):
                ttl = title(si, s)
            else:
                ttl = title

            # get data for show_turbines
            if rotor_color is not None:
                try:
                    turb_angle = self.fres[FV.AMB_WD][si] + self.fres[FV.YAWM][si]
                except KeyError:
                    turb_angle = self.fres[FV.AMB_WD][si]

                show_rotor_dict = {
                    "color": rotor_color,
                    "D": self.fres[FV.D][si],
                    "H": self.fres[FV.H][si],
                    "X": self.fres[FV.X][si],
                    "Y": self.fres[FV.Y][si],
                    "AMB_WD": self.fres[FV.AMB_WD][si],
                    "turb_angle": turb_angle,
                }
            else:
                show_rotor_dict = None

            out = get_fig(
                var=var,
                fig=fig,
                figsize=figsize,
                ax=ax,
                data=data[..., vi],
                si=si,
                s=s,
                levels=levels,
                x_pos=x_pos,
                y_pos=z_pos,
                cmap=cmap,
                xlabel=xlabel,
                ylabel=zlabel,
                title=ttl,
                add_bar=add_bar,
                vlabel=vlabel,
                quiv=quiv,
                vmin=vmin,
                vmax=vmax,
                ret_state=ret_state,
                ret_im=ret_im,
                animated=animated,
                show_rotor_dict=show_rotor_dict,
            )

            yield out

    def gen_states_fig_yz(
        self,
        var,
        x_direction=270.0,
        ylabel="y [m]",
        zlabel="z [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        animated=False,
        rotor_color=None,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a vertical yz-plane.

        Parameters
        ----------
        var: str
            The variable name
        x_direction: float, optional
            The direction of the x axis, 0 = north
        ylabel: str, optional
            The y axis label
        zlabel: str, optional
            The z axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at ech `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        animated: bool
            Switch for usage for an animation
        rotor_color: str, optional
            Indicate the rotor orientation by a colored line
        kwargs: dict, optional
            Additional parameters for SliceData.get_states_data_yz

        Yields
        ------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet

        """
        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        data, states, gdata = self.get_states_data_yz(
            variables=variables,
            vmin={var: vmin} if vmin is not None else {},
            vmax={var: vmax} if vmax is not None else {},
            data_format="numpy",
            ret_states=True,
            ret_grid=True,
            x_direction=x_direction,
            **kwargs,
        )
        x_pos, y_pos, z_pos, __ = gdata

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[..., wdi],
                data[..., wsi],
            )
        )

        # loop over states:
        for si, s in enumerate(states):
            if animated and si > 0 and vmin is not None and vmax is not None:
                add_bar = False
            if not animated and title is None:
                ttl = f"State {s}" if title is None else title
                ttl += f", x direction = {x_direction}°"
                ttl += f", x =  {int(np.round(x_pos))} m"
            elif callable(title):
                ttl = title(si, s)
            else:
                ttl = title

            # get data for show_turbines
            if rotor_color is not None:
                try:
                    turb_angle = self.fres[FV.AMB_WD][si] + self.fres[FV.YAWM][si]
                except KeyError:
                    turb_angle = self.fres[FV.AMB_WD][si]

                show_rotor_dict = {
                    "color": rotor_color,
                    "D": self.fres[FV.D][si],
                    "H": self.fres[FV.H][si],
                    "X": self.fres[FV.X][si],
                    "Y": self.fres[FV.Y][si],
                    "AMB_WD": self.fres[FV.AMB_WD][si],
                    "turb_angle": turb_angle,
                }
            else:
                show_rotor_dict = None

            out = get_fig(
                var=var,
                fig=fig,
                figsize=figsize,
                ax=ax,
                data=data[..., vi],
                si=si,
                s=s,
                levels=levels,
                x_pos=y_pos,
                y_pos=z_pos,
                cmap=cmap,
                xlabel=ylabel,
                ylabel=zlabel,
                title=ttl,
                add_bar=add_bar,
                vlabel=vlabel,
                vmin=vmin,
                vmax=vmax,
                quiv=quiv,
                ret_state=ret_state,
                ret_im=ret_im,
                invert_axis="x",
                animated=animated,
                show_rotor_dict=show_rotor_dict,
            )

            yield out



from foxes.algorithms.sequential import SequentialPlugin

from .flow_plots import FlowPlots2D


class SeqFlowAnimationPlugin(SequentialPlugin):
    """
    Plugin for creating data for a 2D flow animation
    during sequential iterations

    Attributes
    ----------
    orientation: str
        The orientation, either "yx", "xz" or "yz"
    runner: foxes.utils.runners.Runner
        The runner
    pars: dict
        Additional parameters for plotting

    :group: output.flow_plots_2d

    """

    def __init__(self, orientation, runner=None, **pars):
        """
        Constructor.

        Parameters
        ----------
        orientation: str
            The orientation, either "yx", "xz" or "yz"
        runner: foxes.utils.runners.Runner, optional
            The runner
        pars: dict, optional
            Additional parameters for plotting

        """
        super().__init__()
        self.orientation = orientation
        self.runner = runner
        self.pars = pars

        if "title" in self.pars and callable(self.pars["title"]):
            self._tfun = self.pars.pop("title")
        else:
            self._tfun = None

    def initialize(self, algo):
        """
        Initialize data based on the intial iterator

        Parameters
        ----------
        algo: foxes.algorithms.sequential.Sequential
            The current sequetial algorithm

        """
        super().initialize(algo)
        self._data = []

    def update(self, algo, fres, pres=None):
        """
        Updates data based on current iteration

        Parameters
        ----------
        algo: foxes.algorithms.sequential.Sequential
            The latest sequetial algorithm
        fres: xarray.Dataset
            The latest farm results
        pres: xarray.Dataset, optional
            The latest point results

        """
        super().update(algo, fres, pres)

        o = FlowPlots2D(algo, fres, self.runner)

        if self._tfun is not None:
            self.pars["title"] = self._tfun(algo.states.counter, algo.states.index()[0])

        if self.orientation == "xy":
            self._data.append(next(o.gen_states_fig_xy(**self.pars)))
        elif self.orientation == "xz":
            self._data.append(next(o.gen_states_fig_xz(**self.pars)))
        elif self.orientation == "yz":
            self._data.append(next(o.gen_states_fig_yz(**self.pars)))
        else:
            raise KeyError(
                f"Unkown orientation '{self.orientation}', choises: xy, xz, yz"
            )

        if (
            self.pars.get("vmin", None) is not None
            and self.pars.get("vmax", None) is not None
        ):
            self.pars["add_bar"] = False

    def gen_images(self):
        for d in self._data:
            yield d
